// Paul Houle (paulhoule.com) March 22, 2008
//
//  This file is part of par2cmdline (a PAR 2.0 compatible file verification and
//  repair tool). See http://parchive.sourceforge.net for details of PAR 2.0.
//
//  par2cmdline is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  par2cmdline is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "par2cmdline.h"

#pragma warning(disable:4731)		// inhibit VC's "ebp modified" warning
#pragma warning(disable:4799)		// inhibit VC's "missing emms" warning

int PrCap = 0;
int PrNum = 0;

// Sets PrCap to 1 for non-MMX processor, 2 for MMX processor.
// Sets PrNum to number of processors.
// Only needs to be called once immediately after parsing command line.

void SetPr(const CommandLine &commandline) {
	SYSTEM_INFO si;

	if (PrCap) return;

  _asm {
	pushfd								;try to flip "ID" bit in eflags
	pop     eax
	btc     eax,21
	push    eax
	popfd
	pushfd
	pop		ecx
	xor     ecx,eax						;bit 21 == 0 if "ID" bit changed
	xor		eax,eax
	bt      ecx,21
	jc		fset						;jmp= no CPUID, no advanced opcodes
	cpuid								;eax= highest supported CPUID function
	test	eax,eax
	jz		fset						;jmp= can't get features
	mov		eax,1						;get features
	cpuid
	xor		eax,eax
	test	edx,1						;floating point?
	jz		fset						;jmp= no, no MMX
	bt		edx,23						;MMX support?
	jnc		fset						;jmp= nope
	inc		eax							;indicate MMX
fset:
	inc		eax
	mov		PrCap,eax
;$$20150928fdone:
  }

	if (PrCap < 2) {					// if non-MMX,
		if (commandline.GetNoiseLevel() > CommandLine::nlQuiet) {
			cout << "*** Warning: non-MMX processor ***" << endl;
		}
	} else {							// else MMX, allow mult processors
		GetSystemInfo(&si);				// ok for Win95+, all WinNT
		PrNum = si.dwNumberOfProcessors; // fetch # of processors
	}
}

/*----------------------------------------------------------------------------
;
; Builds the combined multiplication tables at the specified address (pMtab).
;
;---------------------------------------------------------------------------*/

void BuildMtab(unsigned *pMtab, unsigned factor, void *table) {
  _asm {
	mov		eax,factor
	movzx	ebx,ah
	movzx	eax,al
	shl		ebx,8
	shl		eax,8
	mov		edx,table
	lea		esi,[edx+eax*2]					;LL
	lea		ecx,[edx+ebx*2+(1*256*256)*2]	;HL
	lea		edi,[edx+eax*2+(1*256*256)*2]	;LH
	lea		edx,[edx+ebx*2+(2*256*256)*2]	;HH

	xor		ebx,ebx
	pxor	mm7,mm7
	mov		eax,pMtab
	align	16
LHmxlp:
	movq	mm0,[esi+ebx]
	pxor	mm0,[ecx+ebx]				;L3,L2,L1,L0
	movq	mm1,mm0
	punpcklwd mm0,mm7
	punpckhwd mm1,mm7
	movq	[eax+ebx*2],mm0
	movq	[eax+ebx*2+8],mm1

	movq	mm2,[edi+ebx]
	pxor	mm2,[edx+ebx]				;H3,H2,H1,H0
	movq	mm3,mm2
	punpcklwd mm2,mm7
	punpckhwd mm3,mm7
	movq	[eax+400h+ebx*2],mm2
	movq	[eax+400h+ebx*2+8],mm3

	add		ebx,8
	cmp		ebx,512
	jne		LHmxlp
  }
}

/*----------------------------------------------------------------------------
;
; Processes block of data a multiple of 4 bytes long.
;
; The amount of data to process (bsize) must be a non-zero multiple of 4.
;
;---------------------------------------------------------------------------*/

void DoBlock4(char *inputbuffer, char *outputbuffer,
	size_t bsize, unsigned *pMtab
) {
  _asm {
	push	ebp
	mov		ebx,bsize					;bytes to process (multiple of 4)
	mov		esi,inputbuffer				;source
	mov		edi,outputbuffer			;destination
	mov		ebp,pMtab					;combined multiplication table

lp4:
	mov		eax,[esi]
	add		esi,4
	mov		ecx,eax
	shr		eax,16
	movzx	edx,al
	shr		eax,8
	mov		edx,[ebp+edx*4]				;look up 2nd solution
	xor		edx,[ebp+400h+eax*4]
	movzx	eax,ch
	movzx	ecx,cl
	shl		edx,16
	mov		eax,[ebp+400h+eax*4]		;1st solution
	xor		eax,[ebp+ecx*4]
	xor		edx,eax						;combined solution
	xor		edx,[edi]
	mov		[edi],edx
	add		edi,4
	sub		ebx,4						;adjust src/dst count, zero if done
	jnz		lp4

	pop		ebp
  }
}

/*----------------------------------------------------------------------------
;
; Processes block of data a multiple of 8 bytes long using SIMD (mmx) opcodes.
;
; It is the caller's responsibility to insure the in/out pointers are
; optimally aligned; if not, performance is reduced significantly.
;
; The amount of data to process (bsize) must be a non-zero multiple of 8.
;
; Unfortunately this routine "reads ahead" 8 bytes beyond the true end of
; the given source block.  The caller must compensate for this by
; appropriately limiting the size of the source.
;
;---------------------------------------------------------------------------*/

void DoBlock8(char *inputbuffer, char *outputbuffer,
	size_t bsize, unsigned *pMtab
) {
  _asm {
	push	ebp
	mov		ebx,bsize					;bytes to process (multiple of 4)
	mov		esi,inputbuffer				;source
	mov		edi,outputbuffer			;destination
	mov		ebp,pMtab					;combined multiplication table

	mov		eax,[esi]					;load 1st 8 source bytes
	movd	mm4,[esi+4]

	add		esi,ebx						;point to end of input/output
	add		edi,ebx
	neg		ebx							;convert byte size to count-up

	; The below loop processes 8 bytes at a time.  It runs at approx.
	; 10.5 clocks per iteration on a Core 2 Duo.
	; That's close to optimal -- which is 10 -- since there are 10 mem
	; reads in the loop (there's only one mem read port in the
	; architecture).  The extra .5 clock appears to be a
	; cache-loading delay, since if the read addresses below are made
	; invariant it goes away.
	;
	; The 8 byte MMX loop runs 20-30% faster than a simple (non-MMX) four
	; byte loop on a Core 2 Duo.  The speedup is really only due to
	; it doing wider memory fetches (8 bytes per src/dst instead of 4).
	;
	; Some work was done to achieve U/V pipe pairing for early Pentiums,
	; but not at the expense of the Core 2 Duo performance.

	align	16
lp8:
	movzx	edx,al
	movzx	ecx,ah
	shr		eax,16
	movd	mm0,[ebp+edx*4]
	movd	mm1,[ebp+400h+ecx*4]
	movzx	edx,al
	movzx	ecx,ah
	movd	eax,mm4
	movq	mm4,[esi+ebx+8]				;read-ahead next 8 source bytes
	movd	mm2,[ebp+edx*4]
	movzx	edx,al
	movq	mm5,[edi+ebx]
	movd	mm3,[ebp+400h+ecx*4]
	movzx	ecx,ah
	shr		eax,16
	punpckldq mm0,[ebp+edx*4]
	movzx	edx,al
	punpckldq mm1,[ebp+400h+ecx*4]
	movzx	ecx,ah
	punpckldq mm2,[ebp+edx*4]
	pxor	mm1,mm0
	punpckldq mm3,[ebp+400h+ecx*4]
	pxor	mm3,mm2
	pslld	mm3,16
	movd	eax,mm4						;prepare src bytes 3-0 for next loop
	pxor	mm1,mm5
	psrlq	mm4,32						;align src bytes 7-4 for next loop
	pxor	mm1,mm3
	movq	[edi+ebx],mm1
	add		ebx,8
	jnz		lp8

	pop		ebp
  }
}

/*----------------------------------------------------------------------------
;
; Threading logic (init, release control struct) and thread function itself.
;
;---------------------------------------------------------------------------*/

#define MAXTH 32						// maximum threads (processors)

typedef struct {						// RS threading control struct
	unsigned init;						// non-zero after struct initialized

	u32 datapresent;					//	invariant (per inputindex) fields
	u32 datamissing;					//	.			.			.
	Galois16 *leftmatrix;				//	.			.			.
	Galois16 *table;					//	.			.			.
	size_t blocklength;					//	.			.			.
	char *inputbuffer;					//	.			.			.
	u32 inputindex;						//	.			.			.
	char *outputbuffer;					//	.			.			.
	size_t chunksize;					//	.			.			.
	u32 totalblockcount;				//	.			.			.

	unsigned nThreads;					// created threads (# cores - 1)

	unsigned blksrepdone;				// blocks reported done
				// volatile fields (may be shared among multiple threads)
	long volatile bltstart;				// blocks left to start
	long volatile bltfinish;			// blocks left to finish

				// kernel objects
	HANDLE sem;							// semaphore used to dispatch threads
	HANDLE ev;							// event used to sense completion
	HANDLE tH[MAXTH];					// vector of thread handles
} RSTH;
RSTH Th = {0};

// This routine processes 1 combined multiplication table and the block it
// applies to.  This is the atomic unit of work done by a thread.
//
// The *pTh structure must be initialized with the invariant fields (only
// the outputindex varies across the default 100 blocks).

void DoBlock(RSTH * const pTh, u32 outputindex
) {
	char Mtab_space[256*2*4+15];		// mult tables (+ extra for alignment)
	unsigned *pMtab;					// ptr to true start of Mtab
	size_t bsize;						// number of bytes to process
	char *inbuf;						// working input  block ptr
	char *outbuf;						// working output block ptr

	// Look up the appropriate element in the RS matrix

	Galois16 factor = pTh->leftmatrix[
	  outputindex * (pTh->datapresent + pTh->datamissing) + pTh->inputindex];
	// Do nothing if the factor happens to be 0
	if (factor == 0) return;

	pMtab = (unsigned *)				// para align true Mtab start addr
		((unsigned)(Mtab_space + 15) & ~15);

	BuildMtab(pMtab, factor, pTh->table); // build Mtab (combined mult tables)

	inbuf = pTh->inputbuffer;			// init in/out addr's
	outbuf = pTh->outputbuffer + (pTh->chunksize * outputindex);
	bsize = pTh->blocklength;			// init bytes to process

	for (;;) {

		// Handle tiny blocks with straightforward 4-byte routine -- for such
		// blocks it's faster to avoid the 8-byte routine's setup overhead.
		//
		// (the "40" number currently below was determined by time testing).

		if (bsize <= 40) {				// if too small to get fancy with,
			if (bsize) {				// if something at all,
				DoBlock4(inbuf, outbuf, bsize, pMtab);
			}
			break;
		}

		// inbuf & outbuf are always relatively 8-byte aligned, but the caller
		// operates in dword sizes so they may not be absolutely 8-byte
		// aligned.  Misalignment slows things down quite a bit (10%); so, if
		// addresses misaligned, process a dword beforehand to align them.

		if ((int) outbuf & 4) {			// if out misaligned (in will be too),
			DoBlock4(inbuf, outbuf, 4, pMtab);
			inbuf += 4;
			outbuf += 4;
			bsize -= 4;
		}

		// Process main portion of the data.  Two details: (1) the DoBlock8
		// code reads ahead 8-bytes, so we limit it to not read beyond the
		// inputbuffer; (2) there may be a residual dword at the end.

		{
			size_t bsz = bsize & ~4;	// size must be a multiple of 8
			bsz -= 8;					// compensate for read-ahead
			DoBlock8(					// process large chunk 8-bytes a shot
				inbuf, outbuf, bsz, pMtab);
			DoBlock4(					// process residual 8-12 bytes
				inbuf + bsz, outbuf + bsz, bsize - bsz, pMtab);
		}
		break;
	}
}

void RelRSthrd(RSTH * const pTh) {	// release RS threading resources
	/*$$20150928 int */ unsigned i;
	
	if (pTh->nThreads) {				// if thread(s) were created,
		pTh->bltstart = -999999;		// command all threads to exit
		ReleaseSemaphore(pTh->sem, pTh->nThreads, NULL);
		WaitForMultipleObjects(			// wait for all threads to exit
			pTh->nThreads, pTh->tH, TRUE, 5000);
		for (i = 0; i < pTh->nThreads; i++) {
			CloseHandle(pTh->tH[i]);	// release thread handles
		}
	}
	if (pTh->ev) CloseHandle(pTh->ev);	// close event & semaphore handles
	if (pTh->sem) CloseHandle(pTh->sem);
}

// Note: the main thread function (below) and the work function it calls
// do not use any C run-time functions; hence, we can use CreateThread
// instead of beginthreadex.

DWORD WINAPI ThreadFunc(LPVOID targ) {	// top-level thread function
	RSTH * const pTh = (RSTH * const) targ;
	long bleft;

	do {
		WaitForSingleObject(pTh->sem, INFINITE); // await triggering
		do {
			bleft =						// number of blocks left to start
				InterlockedDecrement(&pTh->bltstart);
			if (bleft < 0) break;

			DoBlock(pTh,				// process the block
				pTh->totalblockcount - (bleft + 1));
			bleft =						// number of blocks left to finish
				InterlockedDecrement(&pTh->bltfinish);
		} while (bleft);
		if (bleft == 0) {				// if finished last block of set,
			SetEvent(pTh->ev);			// need to wake up the main thread
		}
	} while (bleft > -999999);			// operate till commanded to exit

	return 0;
}

void DoRelRSthrd(void) {				// needed for atexit() in InitRSthrd
	RelRSthrd(&Th);
}

// Create one thread per available extra core.
// If multi-processor operation is not possible sets pTh.nThreads == 0.

void InitRSthrd(RSTH * const pTh) {		// init threading resources
	int i, j;
	long volatile iitest = 99;			// for InterlockedIncrement test

	i = PrNum;							// # of processors
	--i;								// create 1 thread per extra processor
	for (;;) {
		if (i <= 0) break;				// need at least two processors
		// need an OS w/modern InterlockedIncrement (> Win 95 & NT 3.51)
		if (InterlockedIncrement(&iitest) != 100) break;

		i = (i <= MAXTH) ? i : MAXTH;	// limit to our max processors
		pTh->sem = CreateSemaphore(		// semaphore to dispatch threads
			NULL, 0, i * 2, NULL);		//	(non-signaled)
		if (!pTh->sem) break;
		pTh->ev = CreateEvent(			// event to sense end of processing
			NULL, FALSE, FALSE, NULL);	//	(non-signaled, auto-reset)
		if (!pTh->ev) break;
		for (j = 0; j < i; j++) {		// create threads for extra processors
			DWORD ThreadID;
			pTh->tH[j] = CreateThread(
				NULL, 2048, ThreadFunc, pTh, 0, &ThreadID);
			if (!pTh->tH[j]) break;
			pTh->nThreads++;
		}
		break;
	}
	if (iitest == 100) {				// if we tried to create threads,
		if (i == pTh->nThreads) {		//	and were entirely successful,
			atexit(DoRelRSthrd);		// install at-exit cleanup routine
		} else {						// else we tried and failed,
			RelRSthrd(pTh);				// kill anything halfway done
			pTh->nThreads = 0;			// disable threading
		}
	}
}

/*----------------------------------------------------------------------------
;
; Paul Houle (paulhoule.com) March 22, 2008
; The below routine is the gist of the changes.
; It runs the assembly code to handle the main CPU work of par2 creation,
; dispatching multiple threads (if there are cores to do so).
;
; This routine may not be called again (with outputindex != 0) after it has
; reported all blocks in a set have been processed.
;
;---------------------------------------------------------------------------*/

u32 ReedSolomon<Galois16>::AsmProcess(size_t size, u32 inputindex,
 const void *inputbuffer, u32 outputindex, void *outputbuffer,
 size_t chunksize, u32 totalblockcount
) {
	long bleft;

	// First call of many (typically 100) operating within same buffer?

	if (outputindex == 0) {

		Th.datapresent = datapresent;	// copy invariant fields needed by
		Th.datamissing = datamissing;	//	threads (threads have no access
		Th.leftmatrix = leftmatrix;		//	to the ReedSolomon instance)
		Th.table = glmt->tables;
		Th.blocklength = size;
		Th.inputbuffer = (char *) inputbuffer;
		Th.inputindex = inputindex;
		Th.outputbuffer = (char *) outputbuffer;
		Th.chunksize = chunksize;
		Th.totalblockcount = totalblockcount;

		// Set number of blocks that must be started and finished.

		Th.bltstart = totalblockcount;
		Th.bltfinish = totalblockcount;

		Th.blksrepdone = 0;				// count blocks reported done

		if (!Th.init) {					// if haven't created threads,
			Th.init++;					// only run the init code once
			InitRSthrd(&Th);			// create 1 thread per extra core
		}

		if (Th.nThreads					// if we have extra threads,
		 && totalblockcount > 1) {		//	and block(s) to give them,
			u32 ttr = totalblockcount - 1;
			if (ttr > Th.nThreads) ttr = Th.nThreads;
			ReleaseSemaphore(			// release thread(s) to process blocks
				Th.sem, ttr, NULL);
		}
	}

	bleft =								// number of blocks left to start
		InterlockedDecrement(&Th.bltstart);
	if (bleft < 0) {					// if nothing left to do,
		WaitForSingleObject(			// wait for last block(s) to finish
			Th.ev, INFINITE);
		bleft = 0;						// indicate all blocks are done
	} else {
		DoBlock(						// got a block to do, process it
			&Th, Th.totalblockcount - (bleft + 1));
		bleft =							// number of blocks left to finish
			InterlockedDecrement(&Th.bltfinish);
	}

	bleft = Th.totalblockcount - bleft;	// total blocks completed
	bleft -= Th.blksrepdone;			// blocks sensed completed this call
	Th.blksrepdone += bleft;			// update total

	_asm emms
	return (u32) bleft;
}
